/*
  This program and the accompanying materials are
  made available under the terms of the Eclipse Public License v2.0 which accompanies
  this distribution, and is available at https://www.eclipse.org/legal/epl-v20.html
  
  SPDX-License-Identifier: EPL-2.0
  
  Copyright Contributors to the Zowe Project.
*/


#include <yaml.h> 
   /* #include <yaml_private.h> */

#include <stdlib.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdint.h>
#include <errno.h>

#ifdef NDEBUG
#undef NDEBUG
#endif
#include <assert.h>

#include "zowetypes.h"
#include "alloc.h"
#include "utils.h"
#include "charsets.h"
#include "json.h"
#include "jsonschema.h"

#ifdef __ZOWE_OS_ZOS

int __atoe_l(char *bufferptr, int leng);
int __etoa_l(char *bufferptr, int leng);

#endif


#if defined(__ZOWE_OS_ZOS)
#  define SOURCE_CODE_CHARSET CCSID_IBM1047
#else
#  define SOURCE_CODE_CHARSET CCSID_UTF_8
#endif


static int convertToNative(char *buf, size_t size) {
#ifdef __ZOWE_OS_ZOS
  return __atoe_l(buf, size);
#endif
  return 0;
}

static int convertFromNative(char *buf, size_t size) {
#ifdef __ZOWE_OS_ZOS
  return __etoa_l(buf, size);
#endif
  return 0;
}

static char *tokenTypeName(yaml_token_type_t type){
  switch (type){
  case YAML_NO_TOKEN: return "NO_TOKEN";
  case YAML_STREAM_START_TOKEN: return "STREAM_START";
  case YAML_STREAM_END_TOKEN: return "STREAM_END";
  case YAML_VERSION_DIRECTIVE_TOKEN: return "VERSION_DIRECTIVE";
  case YAML_TAG_DIRECTIVE_TOKEN: return "TAG_DIRECTIVE";
  case YAML_DOCUMENT_START_TOKEN: return "DOCUMENT_START";
  case YAML_DOCUMENT_END_TOKEN: return "DOCUMENT_END";
  case YAML_BLOCK_SEQUENCE_START_TOKEN: return "BLOCK_SEQUENCE_START";
  case YAML_BLOCK_MAPPING_START_TOKEN: return "BLOCK_MAPPING_START";
  case YAML_BLOCK_END_TOKEN: return "BLOCK_END";
  case YAML_FLOW_SEQUENCE_START_TOKEN: return "FLOW_SEQUENCE_START";
  case YAML_FLOW_SEQUENCE_END_TOKEN: return "FLOW_SEQUENCE_END";
  case YAML_FLOW_MAPPING_START_TOKEN: return "FLOW_MAPPING_START";
  case YAML_FLOW_MAPPING_END_TOKEN: return "FLOW_MAPPING_END";
  case YAML_BLOCK_ENTRY_TOKEN: return "BLOCK_ENTRY";
  case YAML_FLOW_ENTRY_TOKEN: return "FLOW_ENTRY";
  case YAML_KEY_TOKEN: return "KEY";
  case YAML_VALUE_TOKEN: return "VALUE";
  case YAML_ALIAS_TOKEN: return "ALIAS";
  case YAML_ANCHOR_TOKEN: return "ANCHOR";
  case YAML_TAG_TOKEN: return "TAG";
  case YAML_SCALAR_TOKEN: return "SCALAR";
  default:
    {
      printf("unknown token seen with type = %d\n",(int)type);
      return "UNKNOWN_TOKEN";
    }
  }
}

static char *getScalarStyleName(yaml_scalar_style_t style){
  switch (style){
  case YAML_ANY_SCALAR_STYLE: return "any";
  case YAML_PLAIN_SCALAR_STYLE: return "plain";
  case YAML_SINGLE_QUOTED_SCALAR_STYLE: return "single";
  case YAML_DOUBLE_QUOTED_SCALAR_STYLE: return "double";
  case YAML_LITERAL_SCALAR_STYLE: return "literal";
  case YAML_FOLDED_SCALAR_STYLE: return "folded";
  default: return "UNKNOWN_STYLE";
  }
}

static int yamlReadHandler(void *data, unsigned char *buffer, size_t size, size_t *sizeRead) {
  FILE *fp = data;
  int rc = 1;
  size_t bytesRead = fread(buffer, 1, size, fp);
  if (bytesRead > 0) {
    if (convertFromNative((char *)buffer, bytesRead) == -1) {
    fprintf (stderr, "failed to convert yaml input - %s\n", strerror(errno));
      rc = 0;
    }
  }
  if (ferror(fp)) {
    fprintf (stderr, "failed to read yaml input - %s\n", strerror(errno));
    rc = 0;
  }
  *sizeRead = bytesRead;
  return rc;
}

static void decodeParserError(yaml_parser_t *parser, char *errorBuf, size_t errorBufSize) {
  switch (parser->error) {
    case YAML_MEMORY_ERROR:
      snprintf(errorBuf, errorBufSize, "YAML memory error: not enough memory for parsing");
      break;
    case YAML_READER_ERROR: {
      size_t problemLen = strlen(parser->problem);
      char problemNative[problemLen + 1];
      snprintf (problemNative, problemLen + 1, "%s", parser->problem);
      convertToNative(problemNative, problemLen);
      if (parser->problem_value != -1) {
        snprintf(errorBuf, errorBufSize, "YAML reader error: %s: #%X at %ld", problemNative, parser->problem_value, (long)parser->problem_offset);
      } else {
        snprintf(errorBuf, errorBufSize, "YAML reader error: %s at %ld", problemNative, (long)parser->problem_offset);
      }
      break;
    }
    case YAML_SCANNER_ERROR:
      if (parser->context) {
        snprintf(errorBuf, errorBufSize, "YAML scanner error: %s at line %d, column %d"
                "%s at line %d, column %d\n", parser->context,
                (int)parser->context_mark.line+1, (int)parser->context_mark.column+1,
                parser->problem, (int)parser->problem_mark.line+1,
                (int)parser->problem_mark.column+1);
      } else {
        snprintf(errorBuf, errorBufSize, "YAML scanner error: %s at line %d, column %d",
                 parser->problem, (int)parser->problem_mark.line+1,
                 (int)parser->problem_mark.column+1);
      }
      break;
      case YAML_PARSER_ERROR:
        if (parser->context) {
          snprintf(errorBuf, errorBufSize, "YAML parser error: %s at line %d, column %d\n"
                   "%s at line %d, column %d", parser->context,
                   (int)parser->context_mark.line+1, (int)parser->context_mark.column+1,
                   parser->problem, (int)parser->problem_mark.line+1,
                   (int)parser->problem_mark.column+1);
        } else {
          snprintf(errorBuf, errorBufSize, "YAML parser error: %s at line %d, column %d",
                   parser->problem, (int)parser->problem_mark.line+1,
                   (int)parser->problem_mark.column+1);
        }
        break;
      default:
        snprintf(errorBuf, errorBufSize, "YAML parser: unknown error");
  }
}

yaml_document_t *readYAML(const char *filename, char *errorBuf, size_t errorBufSize) {
  FILE *file = NULL;
  yaml_document_t *document = NULL;
  yaml_parser_t parser = {0};
  bool done = false;

  do {
    if (!(document = (yaml_document_t*)safeMalloc(sizeof(yaml_document_t), "YAML Doc"))) {
      snprintf(errorBuf, errorBufSize, "failed to alloc memory for YAML doc");
      break;      
    }
    memset(document, 0, sizeof(yaml_document_t));
    if (!(file = fopen(filename, "rb"))) {
      snprintf(errorBuf, errorBufSize, "failed to read '%s' - %s", filename, strerror(errno));
      break;
    }
    if (!yaml_parser_initialize(&parser)) {
      snprintf(errorBuf, errorBufSize, "failed to initialize YAML parser");
      break;
    }
    yaml_parser_set_input(&parser, yamlReadHandler, file);
    if (!yaml_parser_load(&parser, document)) {
      decodeParserError(&parser, errorBuf, errorBufSize);
      break;
    }
    if (!yaml_document_get_root_node(document)){
      snprintf(errorBuf, errorBufSize, "failed to get root node in YAML '%s'", filename);
      break;
    }
    done = true;
  } while (0);

  if (!done && document) {
    yaml_document_delete(document);
    safeFree((char*)document, sizeof(yaml_document_t));
    document = NULL;
  }
  if (file) {
    fclose(file);
  }
  yaml_parser_delete(&parser);
  return document;
}

static void indent(int x){
  for (int i=0; i<x; i++){
    printf(" ");
  }
}

static char *getSequenceStyleName(yaml_sequence_style_t style){
  switch (style){
  case YAML_ANY_SEQUENCE_STYLE: return "any";
  case YAML_BLOCK_SEQUENCE_STYLE: return "block";
  case YAML_FLOW_SEQUENCE_STYLE: return "flow";
  default: return "unknown";
  }
}

#define SCALAR_SIZE_LIMIT 1024

static void printYamlScalar(const yaml_node_t *node, bool eol) {
  size_t dataLen = node->data.scalar.length;
  int printLength = dataLen > SCALAR_SIZE_LIMIT ? 40 : (int)dataLen;
  char val[printLength + 1];
  snprintf(val, printLength + 1, "%.*s", printLength, node->data.scalar.value);
  convertToNative(val, printLength);
  printf("%s%s", val, eol ? "\n" : "");
}


static void pprintYAML1(yaml_document_t *doc, yaml_node_t *node, int depth){
  switch (node->type){
  case YAML_NO_NODE:
    {
      indent(depth);
      printf("NoNode\n");
    }
    break;
  case YAML_SCALAR_NODE:
    {
      indent(depth);
      size_t dataLen = node->data.scalar.length;
      printf("Scalar: (len=%d)", (int)node->data.scalar.length);
      printYamlScalar(node, true);
    }
    break;
  case YAML_SEQUENCE_NODE:
    {
      yaml_node_item_t *item;
      indent(depth);printf("SequenceStart (%s)\n",getSequenceStyleName(node->data.sequence.style));
      for (item = node->data.sequence.items.start; item < node->data.sequence.items.top; item++) {
        yaml_node_t *nextNode = yaml_document_get_node(doc, *item);
        if (nextNode){
          pprintYAML1(doc,nextNode,depth+2);
        } else {
          indent(depth+4);
          printf("dead end item\n");
        }
      }
      indent(depth);printf("SequenceEnd\n");
    }
    break;
  case YAML_MAPPING_NODE:
   {
    yaml_node_pair_t *pair;
    indent(depth);printf("MapStart\n");
    for (pair = node->data.mapping.pairs.start; pair < node->data.mapping.pairs.top; pair++) {
      yaml_node_t *keyNode = yaml_document_get_node(doc, pair->key);
      /* printf("keyNode 0x%p\n",keyNode); */
      if (keyNode) {
        if (keyNode->type != YAML_SCALAR_NODE){
          printf("*** UNEXPECTED NON SCALAR KEY ***\n");
        } else {
          indent(depth+2);
          size_t dataLen = keyNode->data.scalar.length;
          int printLength = dataLen > SCALAR_SIZE_LIMIT ? 40 : (int)dataLen;
          /* printf("printLength=%d dataLen=%lld\n",printLength,dataLen); */
          printYamlScalar(keyNode, false);
          printf(":\n");
        }
      } else {
        printf("dead end key\n");
      }
      yaml_node_t *valueNode = yaml_document_get_node(doc, pair->value);
      if (valueNode){
        pprintYAML1(doc,valueNode,depth+4);
      } else{
        printf("dead end value\n");
      }
    }
    indent(depth);printf("MapEnd\n");
  }
    break;
  default:
    printf("unexpected yaml node type %d\n",node->type);
    break;
  }
}
  
void pprintYAML(yaml_document_t *document){
  pprintYAML1(document,yaml_document_get_root_node(document),0);
}

#define MAX_JSON_KEY 256
#define MAX_JSON_STRING 65536

/* this needs to be smarter some day
   Refer to Yaml spec http://yaml.org/spec/1.2-old/spec.html#id2805071
 */

static bool isSyntacticallyInteger(yaml_char_t *data, int length){
  if (length == 0){
    /* printf("empty string not integer\n"); */
    return false;
  }
  for (int i=0; i<length; i++){
    yaml_char_t c = data[i];
    if (c >= '0' && c <= '9'){
      
    } else {
      /* printf("%s is NOT an integer\n",data); */
      return false;
    }     
  }
  /* printf("%s is an integer\n",data); */
  return true;
}

static bool isSyntacticallyBool(yaml_char_t *data, int length){
  if ((length == 4) &&
      (!memcmp(data,"true",4) ||
       !memcmp(data,"True",4) ||
       !memcmp(data,"TRUE",4))){
    return true;
  } else if ((length == 5) &&
             (!memcmp(data,"false",5) ||
              !memcmp(data,"False",5) ||
              !memcmp(data,"FALSE",5))){
    return true;
  } else {
    return false;
  }
}

static bool isSyntacticallyNull(yaml_char_t *data, int length){
  if ((length == 1) &&
      !memcmp(data,"~",1)){
    return true;
  } else if ((length == 4) &&
             (!memcmp(data,"null",4) ||
              !memcmp(data,"Null",4) ||
              !memcmp(data,"NULL",4))){
    return true;
  } else {
    return false;
  }
}

static bool isSyntacticallyTemplate(yaml_char_t *data, int length){
  if (strstr((char*)data,"${{")){
    return true;
  } else {
    return false;
  }
}

static int64_t readInt(yaml_char_t *data, int length, bool *valid){
  int64_t val64 = 0;
  bool allDecimal = true;
  
  for (int i=0; i<length; i++){
    yaml_char_t c = data[i];
    if (c >= '0' && c <= '9'){
      val64 = (10 * val64) + (c-'0');
    } else {
      allDecimal = false;
    }
  }

  *valid = allDecimal;
  return val64;
}

static bool readBool(yaml_char_t *data, int length, bool *valid){
  if ((length == 4) &&
      (!memcmp(data,"true",4) ||
       !memcmp(data,"True",4) ||
       !memcmp(data,"TRUE",4))){
    *valid = true;
    return true;
  } else if ((length == 5) &&
             (!memcmp(data,"false",5) ||
              !memcmp(data,"False",5) ||
              !memcmp(data,"FALSE",5))){
    *valid = true;
    return false;
  } else {
    *valid = false;
    return false;
  }
}


#define JSON_FAIL_NOT_HANDLED  100
#define JSON_FAIL_BAD_INTEGER  104
#define JSON_FAIL_BAD_BOOL     108
#define JSON_FAIL_BAD_NULL     112
#define JSON_FAIL_BAD_TEMPLATE 116

static char *extractString(JsonBuilder *b, char *s, char *e){
  int len = e-s;
  char *copy = SLHAlloc(b->parser.slh,len+1);
  memcpy(copy,s,len);
  copy[len] = 0;
  return copy;
}

static void addPlusIfNecessary(jsonPrinter *p, int sourceCCSID, bool *firstPtr){
  bool first = *firstPtr;
  if (!first){
    jsonWriteParseably(p," + ",3,false,false,sourceCCSID);
  }
  *firstPtr = false;
}

static int buildTemplateJSON(JsonBuilder *b, Json *parent, char *parentKey,
                             char *nativeValue, int valueLength){
  char *tail = nativeValue;
  JsonBuffer *buffer = makeJsonBuffer();
  int sourceCCSID = CCSID_UTF_8;
  /* printf ("buildTemplateJSON nativeValue '%.*s' sourceCodeCharset %d\n", valueLength, nativeValue, sourceCCSID);*/
  jsonPrinter *p = makeBufferJsonPrinter(sourceCCSID,buffer);
  int status = 0;
  bool first = true;
  while (true){
    char *nextExpr = strstr(tail,"${{");
    
    if (nextExpr){
      if (nextExpr > tail){
        char *frag = extractString(b,tail,nextExpr);
        /* printf("frag = '%s'\n",frag); */
        addPlusIfNecessary(p,sourceCCSID,&first);
        jsonWriteParseably(p,frag,strlen(frag),true,false,sourceCCSID);
      }
      char *end = strstr(nextExpr,"}}");
      if (end){
        char *exprText = extractString(b,nextExpr+3,end);
        tail = end+2;
        addPlusIfNecessary(p,sourceCCSID,&first);
        jsonWriteParseably(p,exprText,strlen(exprText),false,false,sourceCCSID);
      } else {
        status = JSON_FAIL_BAD_TEMPLATE;
        break;
      }
    } else {
      char *lastFrag = extractString(b,tail,nativeValue+valueLength);
      /* printf("lastFrag = '%s'\n",lastFrag); */
      if (strlen(lastFrag) > 0){
        addPlusIfNecessary(p,sourceCCSID,&first);
        jsonWriteParseably(p,lastFrag,strlen(lastFrag),true,false,sourceCCSID);
      }
      break;
    }
  }
  if (status == 0){
    Json *object = jsonBuildObject(b,parent,parentKey,&status);
    if (!status){
      jsonBuildString(b,object,ZOWE_INTERNAL_TYPE,ZOWE_UNEVALUATED,strlen(ZOWE_UNEVALUATED),&status);
      if (!status){
        jsonBufferTerminateString(buffer);
        char *sourceCode = jsonBufferCopy(buffer);
        if (b->traceLevel >= 1){
          printf("embedded source code is (tl=%d): %s\n",b->traceLevel,sourceCode);
          dumpbuffer(sourceCode,strlen(sourceCode));
        }
        jsonBuildString(b,object,"source",sourceCode,strlen(sourceCode),&status);
      }
    }
  }
  freeJsonPrinter(p);
  freeJsonBuffer(buffer);
  return status;
}


static Json *yaml2JSON1(JsonBuilder *b, Json *parent, char *parentKey,
                        yaml_document_t *doc, yaml_node_t *node, int depth){
  int buildStatus = 0;
  switch (node->type){
  case YAML_NO_NODE:
    {
      printf("*** WARNING *** NoNode\n");
      return NULL;
    }
  case YAML_SCALAR_NODE:
    {
      size_t dataLen = node->data.scalar.length;
      int valueLength = (int)dataLen;
      if (dataLen > MAX_JSON_STRING){
        printf("*** WARNING *** oversize JSON string!\n");
        valueLength = MAX_JSON_STRING;
      }
      // Negative numbers, hexadecimal
      yaml_scalar_style_t style = node->data.scalar.style;
      switch (style){
      case YAML_ANY_SCALAR_STYLE:
      case YAML_PLAIN_SCALAR_STYLE:
      case YAML_SINGLE_QUOTED_SCALAR_STYLE:
      case YAML_DOUBLE_QUOTED_SCALAR_STYLE:
      case YAML_LITERAL_SCALAR_STYLE:
      case YAML_FOLDED_SCALAR_STYLE:
      {
        char nativeValue[valueLength+1];
        char *tag = (char*)node->tag;
        int tagLen = strlen(tag);
        char nativeTag[tagLen + 1];
        snprintf(nativeValue, valueLength+1, "%.*s", valueLength, (const char *)node->data.scalar.value);
        snprintf(nativeTag, tagLen + 1, "%.*s", tagLen, tag);
        convertToNative(nativeValue, valueLength);
        convertToNative(nativeTag, tagLen);
        if (b->traceLevel >= 2){
          printf("tag = %s scalarStyle=%s\n",nativeTag,getScalarStyleName(node->data.scalar.style));
        }
        Json *scalar = NULL;
        // HERE, make test with float, int, bool, null, ddate
        if (!strcmp(nativeTag,YAML_NULL_TAG) ||
            (!strcmp(nativeTag,YAML_STR_TAG) &&
             (style == YAML_PLAIN_SCALAR_STYLE) &&
             isSyntacticallyNull((yaml_char_t*)nativeValue,valueLength))){
          scalar = jsonBuildNull(b,parent,parentKey,&buildStatus);
        } else if (!strcmp(nativeTag,YAML_INT_TAG) ||
                   (!strcmp(nativeTag,YAML_STR_TAG) &&
                    (style == YAML_PLAIN_SCALAR_STYLE) &&
                    isSyntacticallyInteger((yaml_char_t*)nativeValue,valueLength))){
          bool valid;
          int64_t x = readInt((yaml_char_t*)nativeValue,valueLength,&valid);
          if (valid){
            scalar = jsonBuildInt64(b,parent,parentKey,x,&buildStatus);
          } else {
            buildStatus = JSON_FAIL_BAD_INTEGER;
          }
        } else if (!strcmp(nativeTag,YAML_BOOL_TAG) ||
                   (!strcmp(nativeTag,YAML_STR_TAG) &&
                    (style == YAML_PLAIN_SCALAR_STYLE) &&
                    isSyntacticallyBool((yaml_char_t*)nativeValue,valueLength))){
          bool valid;
          bool x = readBool((yaml_char_t*)nativeValue,valueLength,&valid);
          if (valid){
            scalar = jsonBuildBool(b,parent,parentKey,x,&buildStatus);
          } else {
            buildStatus = JSON_FAIL_BAD_BOOL;
          }
        } else if (!strcmp(nativeTag,YAML_STR_TAG)){
          /*
            Must be plain
            ${{ <javaScript }} (otherText  ${{ }})*
            how to be integer??
          */
          if ((style == YAML_PLAIN_SCALAR_STYLE) &&
              isSyntacticallyTemplate((yaml_char_t*)nativeValue,valueLength)){
            buildStatus = buildTemplateJSON(b,parent,parentKey,nativeValue,valueLength);
          } else if (style == YAML_SINGLE_QUOTED_SCALAR_STYLE &&
                     isSyntacticallyTemplate((yaml_char_t*)nativeValue,valueLength)){
            buildStatus = JSON_FAIL_NOT_HANDLED;
          } else if (style == YAML_DOUBLE_QUOTED_SCALAR_STYLE &&
                     isSyntacticallyTemplate((yaml_char_t*)nativeValue,valueLength)){
            buildStatus = buildTemplateJSON(b,parent,parentKey,nativeValue,valueLength);
          } else {
            scalar = jsonBuildString(b,parent,parentKey,nativeValue,valueLength,&buildStatus);
          }
        } else if (!strcmp(nativeTag,YAML_FLOAT_TAG)){
          printf("*** Warning don't know how to handle float yet\n");
          buildStatus = JSON_FAIL_NOT_HANDLED;
        } else if (!strcmp(nativeTag,YAML_TIMESTAMP_TAG)){
          printf("*** Warning don't know how to handle timestamp yet\n");
          buildStatus = JSON_FAIL_NOT_HANDLED;
        }
        if (buildStatus){
          printf("*** WARNING *** Failed to add property/scalar err=%d\n",buildStatus);
        }
        return scalar;
      }
      default:
        printf("*** WARNING *** - unknown scalar style %d",node->data.scalar.style);
        return NULL;
      }
    }
  case YAML_SEQUENCE_NODE:
    {
      yaml_node_item_t *item;
      Json *jsonArray = jsonBuildArray(b,parent,parentKey,&buildStatus);
      if (jsonArray){
        for (item = node->data.sequence.items.start; item < node->data.sequence.items.top; item++) {
          yaml_node_t *nextNode = yaml_document_get_node(doc, *item);
          if (nextNode){
            yaml2JSON1(b,jsonArray,NULL,doc,nextNode,depth+2);
          } else {
            printf("*** WARNING *** dead end item\n");
          }
        }
        return jsonArray;
      } else {
        printf("*** WARNING *** Failed to add json array err=%d\n",buildStatus);
        return NULL;
      }
    }
  case YAML_MAPPING_NODE:
    {
      yaml_node_pair_t *pair;
      Json *jsonObject = jsonBuildObject(b,parent,parentKey,&buildStatus);
      if (jsonObject){
        for (pair = node->data.mapping.pairs.start; pair < node->data.mapping.pairs.top; pair++) {
          yaml_node_t *keyNode = yaml_document_get_node(doc, pair->key);
          /* printf("keyNode 0x%p\n",keyNode); */
          char *key = NULL;
          int keyLength;
          if (keyNode) {
            if (keyNode->type != YAML_SCALAR_NODE){
              printf("*** WARNING *** Non Scalar key\n");
            } else {
              size_t dataLen = keyNode->data.scalar.length;
              keyLength = (int)dataLen;
              if (keyLength > MAX_JSON_KEY){
                printf("*** WARNING *** key too long '%*.*s...'\n",
                       MAX_JSON_KEY,MAX_JSON_KEY,keyNode->data.scalar.value);
              } else {
                key = (char*)keyNode->data.scalar.value;
              }
            }
          } else {
            printf("*** WARNING *** dead end key\n");
          }
          if (key){
            char *keyNative = jsonBuildKey(b, key, keyLength);
            convertToNative(keyNative, keyLength);
            yaml_node_t *valueNode = yaml_document_get_node(doc, pair->value);
            if (valueNode){
              yaml2JSON1(b,jsonObject,keyNative,doc,valueNode,depth+2);
            } else{
              printf("*** WARNING *** dead end value\n");
            }
          }
        }
        return jsonObject;
      } else {
        printf("*** WARNING *** Failed to add json object err=%d\n",buildStatus);
        return NULL;
      }
    }
    break;
  default:
    printf("*** WARNING *** unexpected yaml node type %d\n",node->type);
    return NULL;
  }
}

typedef struct ByteOutputStream_tag{
  int   size;
  int   capacity;
  int   chunkSize;
  char *data;
} ByteOutputStream;

ByteOutputStream *makeByteOutputStream(int chunkSize){
  ByteOutputStream *bos = (ByteOutputStream*)safeMalloc(sizeof(ByteOutputStream),"ByteOutputStream");
  bos->size = 0;
  bos->capacity = chunkSize;
  bos->chunkSize = chunkSize;
  bos->data = safeMalloc(chunkSize,"ByteOutputStream Chunk");
  return bos;
}

int bosWrite(ByteOutputStream *bos, char *data, int dataSize){
  if (bos->size+dataSize > bos->capacity){
    int extendSize = (bos->chunkSize > dataSize) ? bos->chunkSize : dataSize;
    printf("bos extend currSize=0x%x dataSize=0x%x chunk=0x%x extend=0x%x\n",
           bos->size,dataSize,bos->chunkSize,extendSize);
    int newCapacity = bos->capacity + extendSize;
    char *newData = safeMalloc(newCapacity,"BOS extend");
    memcpy(newData,bos->data,bos->size);
    safeFree(bos->data,bos->capacity);
    bos->data = newData;
    bos->capacity = newCapacity;
  }
  memcpy(bos->data+bos->size,data,dataSize);
  bos->size += dataSize;
  return bos->size;
}

int bosAppendString(ByteOutputStream *bos, char *s){
  return bosWrite(bos,s,strlen(s));
}

int bosAppendChar(ByteOutputStream *bos, char c){
  return bosWrite(bos,&c,1);
}

char *bosNullTerminateAndUse(ByteOutputStream *bos){
  char c = 0;
  bosWrite(bos,&c,1);
  return bos->data;
}

char *bosUse(ByteOutputStream *bos){
  return bos->data;
}

void bosReset(ByteOutputStream *bos){
  bos->size = 0;
}

void bosFree(ByteOutputStream *bos, bool freeBuffer){
  if (freeBuffer){
    safeFree(bos->data,bos->capacity);
  }
  safeFree((char*)bos,sizeof(ByteOutputStream));
}

/* Friday, expand it */

Json *yaml2JSON(yaml_document_t *document, ShortLivedHeap *slh){
  JsonBuilder *builder = makeJsonBuilder(slh);
  Json *json = yaml2JSON1(builder,NULL,NULL,document,yaml_document_get_root_node(document),0);
  freeJsonBuilder(builder,false);
  return json;
}


static void yaml2JS1(ByteOutputStream *bos,
                     yaml_document_t *doc, yaml_node_t *node, int depth,
                     int traceLevel){
  int buildStatus = 0;
  switch (node->type){
  case YAML_NO_NODE:
    {
      printf("*** WARNING *** NoNode\n");
      return;
    }
  case YAML_SCALAR_NODE:
    {
      size_t dataLen = node->data.scalar.length;
      int valueLength = (int)dataLen;
      if (dataLen > MAX_JSON_STRING){
        printf("*** WARNING *** oversize JSON string!\n");
        valueLength = MAX_JSON_STRING;
      }
      // Negative numbers, hexadecimal
      yaml_scalar_style_t style = node->data.scalar.style;
      switch (style){
      case YAML_ANY_SCALAR_STYLE:
      case YAML_PLAIN_SCALAR_STYLE:
      case YAML_SINGLE_QUOTED_SCALAR_STYLE:
      case YAML_DOUBLE_QUOTED_SCALAR_STYLE:
      case YAML_LITERAL_SCALAR_STYLE:
      case YAML_FOLDED_SCALAR_STYLE:
      {
        char nativeValue[valueLength+1];
        char *tag = (char*)node->tag;
        int tagLen = strlen(tag);
        char nativeTag[tagLen + 1];
        snprintf(nativeValue, valueLength+1, "%.*s", valueLength, (const char *)node->data.scalar.value);
        snprintf(nativeTag, tagLen + 1, "%.*s", tagLen, tag);
        convertToNative(nativeValue, valueLength);
        convertToNative(nativeTag, tagLen);
        if (traceLevel >= 2){
          printf("tag = %s scalarStyle=%s\n",nativeTag,getScalarStyleName(node->data.scalar.style));
        }
        Json *scalar = NULL;
        /* HERE, make test with float, int, bool, null, date */
        if (!strcmp(nativeTag,YAML_NULL_TAG)){
        } else if (!strcmp(nativeTag,YAML_NULL_TAG)){
          /* Json *scalar = jsonBuildNull(b,parent,parentKey,&buildStatus); */
        } else if (!strcmp(nativeTag,YAML_BOOL_TAG)){
          /* Json *scalar = jsonBuildBool(b,parent,parentKey,"FOO",3,&buildStatus); */
        } else if (!strcmp(nativeTag,YAML_INT_TAG) ||
                   (!strcmp(nativeTag,YAML_STR_TAG) &&
                    (style == YAML_PLAIN_SCALAR_STYLE) &&
                    isSyntacticallyInteger((yaml_char_t*)nativeValue,valueLength))){
          bool valid;
          int64_t x = readInt((yaml_char_t*)nativeValue,valueLength,&valid);
          if (valid){
            /* this is not yet done */
          } else {
            buildStatus = JSON_FAIL_BAD_INTEGER;
          }
          /* Json *scalar = jsonBuildInt(b,parent,parentKey,"FOO",3,&buildStatus); */
        } else if (!strcmp(nativeTag,YAML_STR_TAG)){
          printf("%*.*s\n",valueLength,valueLength,nativeValue);
        } else if (!strcmp(nativeTag,YAML_FLOAT_TAG)){
          printf("*** Warning don't know how to handle float yet\n");
          buildStatus = JSON_FAIL_NOT_HANDLED;
        } else if (!strcmp(nativeTag,YAML_TIMESTAMP_TAG)){
          printf("*** Warning don't know how to handle timestamp yet\n");
          buildStatus = JSON_FAIL_NOT_HANDLED;
        }
        if (buildStatus){
          printf("*** WARNING *** Failed to add property/scalar err=%d\n",buildStatus);
        }
        return;
      }
      default:
        printf("*** WARNING *** - unknown scalar style %d",node->data.scalar.style);
        return;
      }
    }
  case YAML_SEQUENCE_NODE:
    {
      yaml_node_item_t *item;
      printf("[\n");
      for (item = node->data.sequence.items.start; item < node->data.sequence.items.top; item++) {
        yaml_node_t *nextNode = yaml_document_get_node(doc, *item);
        if (nextNode){
          yaml2JS1(bos,doc,nextNode,depth+2,traceLevel);
        } else {
          printf("*** WARNING *** dead end item\n");
        }
      }
      printf("]\n");
    }
  case YAML_MAPPING_NODE:
    {
      yaml_node_pair_t *pair;
      /* Json *jsonObject = jsonBuildObject(b,parent,parentKey,&buildStatus); */
      for (pair = node->data.mapping.pairs.start; pair < node->data.mapping.pairs.top; pair++) {
        yaml_node_t *keyNode = yaml_document_get_node(doc, pair->key);
        /* printf("keyNode 0x%p\n",keyNode); */
        char *key = NULL;
        int keyLength;
        if (keyNode) {
          if (keyNode->type != YAML_SCALAR_NODE){
            printf("*** WARNING *** Non Scalar key\n");
          } else {
            size_t dataLen = keyNode->data.scalar.length;
            keyLength = (int)dataLen;
            if (keyLength > MAX_JSON_KEY){
              printf("*** WARNING *** key too long '%*.*s...'\n",
                     MAX_JSON_KEY,MAX_JSON_KEY,keyNode->data.scalar.value);
            } else {
                key = (char*)keyNode->data.scalar.value;
            }
          }
        } else {
          printf("*** WARNING *** dead end key\n");
        }
        if (key){
          char keyNative[keyLength+1];
          memcpy(keyNative,key,keyLength);
          keyNative[keyLength] = 0;
          convertToNative(keyNative, keyLength);
          yaml_node_t *valueNode = yaml_document_get_node(doc, pair->value);
          if (valueNode){
            printf("%s\n",keyNative);
            yaml2JS1(bos,doc,valueNode,depth+2,traceLevel);
          } else{
            printf("*** WARNING *** dead end value\n");
          }
        }
      }
      return;
    }
    break;
  default:
    printf("*** WARNING *** unexpected yaml node type %d\n",node->type);
    return;
  }
}

/* this is not fully done, and might not be needed */
static void yaml2JS(yaml_document_t *document, ShortLivedHeap *slh){
  ByteOutputStream *bos = makeByteOutputStream(0x1000);
  yaml2JS1(bos,document,yaml_document_get_root_node(document),0,1);
}

/*
  This program and the accompanying materials are
  made available under the terms of the Eclipse Public License v2.0 which accompanies
  this distribution, and is available at https://www.eclipse.org/legal/epl-v20.html
  
  SPDX-License-Identifier: EPL-2.0
  
  Copyright Contributors to the Zowe Project.
*/
